// Copyright (C) 2024
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#![cfg_attr(not(debug_assertions), windows_subsystem = "windows")] // hide console window on Windows in release
#![allow(rustdoc::missing_crate_level_docs)] // it's an example

// -- Uses: ---------------------------------------------------------------
use eframe::egui::{self, Color32};
use std::path::Path;
use std::time::{Duration, Instant};
use std::sync::{Arc, RwLock};

// -- Constants: ----------------------------------------------------------
const MAX_THREADS: u8 = 40;
const IMGFILEPATH: &str = "/home/acorbi/projects/image-pc/images/Ferris.png";

// -- Main program: -------------------------------------------------------
fn main() -> eframe::Result {
    env_logger::init(); // Log to stderr (if you run with `RUST_LOG=debug`).
    let options = eframe::NativeOptions {
        viewport: egui::ViewportBuilder::default().with_inner_size([400.0, 800.0]),
        ..Default::default()
    };

    let app = MyApp::new(1);

    eframe::run_native(
        "Image Viewer",
        options,
        Box::new(|cc| {
            // This gives us image support:
            egui_extras::install_image_loaders(&cc.egui_ctx);
            Ok(Box::<MyApp>::new(app))
        }),
    )
}

// -- Application: --------------------------------------------------------
#[derive(Default)]
struct MyApp {
    texture: Option<egui::TextureHandle>,
    img: Option<egui::ColorImage>,
    nthr: u8,
    etime: Duration,
}

impl eframe::App for MyApp {
    fn update(&mut self, ctx: &egui::Context, _frame: &mut eframe::Frame) {
        let imgfile = IMGFILEPATH;

        egui::CentralPanel::default()
            .show(ctx, |ui| {
                egui::Window::new("Programación Concurrente").show(ctx, |ui| {
                    let abt = egui::RichText::new("Actions").
                        color(egui::Color32::RED).
                        size(18.0);
                    ui.label(abt);

                    ui.add_space(2.0);

                    ui.horizontal(|ui| {
                        let rbt = egui::RichText::new("Reload Image").
                            color(egui::Color32::LIGHT_YELLOW).
                        size(14.0);
                        if ui.add(egui::Button::new(rbt)).clicked() {
                            self.reload_image();
                        }

                        ui.add_space(4.0);

                        let gbt = egui::RichText::new("Grayscale image").
                            color(egui::Color32::LIGHT_YELLOW).
                        size(14.0);
                        if ui.add(egui::Button::new(gbt)).clicked() {
                            self.grayscale_image();
                        }
                    });

                    ui.add_space(8.0);

                    // Nthreads control
                    let ntht = egui::RichText::new("N. threads").
                        color(egui::Color32::LIGHT_RED).
                        size(16.0);
                    ui.add(egui::Slider::new(&mut self.nthr, 1..=MAX_THREADS).text(ntht));

                    ui.add_space(8.0);

                    let etlbl = egui::RichText::new(format!("Elapsed time: {:0.2}s.", self.etime.as_secs_f32())).
                        color(egui::Color32::WHITE).
                        size(18.0);
                    ui.label(etlbl);

                    ui.add_space(8.0);

                    // Quit app
                    let qbt = egui::RichText::new("Quit").
                        color(egui::Color32::GOLD).
                        size(18.0);
                    if ui.button(qbt).clicked() {
                        ctx.send_viewport_cmd(egui::ViewportCommand::Close);
                    }
                });

                egui::ScrollArea::both()
                    .show(ui, |ui| {

                        // 1. Get current image color or grayscale
                        let ferris = self.img.get_or_insert_with (|| {
                            //println!("Loading image...");
                            load_image_from_path(Path::new(imgfile)).unwrap()
                        });

                        // 2. Prepare the texture:
                        let texture: &mut egui::TextureHandle = self.texture.get_or_insert_with(|| {
                            // Load the texture only once.
                            ui.ctx().load_texture("ferris", ferris.clone(), Default::default())
                        });

                        // 3. Show the image:
                        ui.image((texture.id(), texture.size_vec2()));
                    });
            });
    }
}

impl MyApp {
    fn new(nthr: u8) -> Self {
        MyApp {
            nthr,
            ..Default::default()
        }
    }

    fn reload_image (&mut self) {
        let imgfile = IMGFILEPATH;
        self.img = Some(load_image_from_path(Path::new(imgfile)).unwrap());
        // Force texture reload in 'update' method.
        self.texture = None;
        //println!("Image reloaded.");
    }

    fn grayscale_image(&mut self) {
        // Force texture reload in 'update' method.
        self.texture = None;

        // width & height of each new region/portion of original image
        let w = self.img.as_ref().unwrap().width() as f32;
        let h = self.img.as_ref().unwrap().height() as f32 / (self.nthr as f32);
        //println!("GrayScaling: {w}X{h} - {}_nth", self.nthr);

        // Prepare Arc for the colorimage
        let imgcopy = self.img.as_mut().unwrap();
        let arcmtxci = Arc::new(RwLock::new(imgcopy));

        // Time measure: Marca el inicio
        let start = Instant::now();

        // Compute new portions rectangle
        let mut y = 0.0;
        for _p in 0..self.nthr {
            let min: egui::Pos2 = [0.0, y].into();
            let size: egui::Vec2 = [w, h].into();
            let r : egui::Rect = egui::Rect::from_min_size(min, size);
            y += h;

            // Create threads
            let imgclone = arcmtxci.clone();

            std::thread::scope(|s| {
                // Thread id vector
                //let mut tidv = vec![];

                let _thid = s.spawn (move || {
                    MyApp::grayscale_portion(imgclone, r);
                });

                // tidv.push(thid);
                // for t in tidv {
                //     let _ = t.join();
                // }
            });
        }


        // Calcula el tiempo transcurrido
        self.etime = start.elapsed();
    }

    // Class method
    fn grayscale_portion(arcci: Arc<RwLock<&mut egui::ColorImage>>, r: egui::Rect) {
        let sx = r.min.x as usize;
        let sy = r.min.y as usize;
        let w = r.max.x as usize;
        let h = r.max.y as usize;

        // let nanosleep = std::time::Duration::from_nanos(10);
        // thread::sleep(nanosleep);

        for x in sx..w {
            for y in sy..h {
                let img = arcci.read().unwrap();
                let r = img[(x,y)].r();
                let g = img[(x,y)].g();
                let b = img[(x,y)].b();
                let a = img[(x,y)].a();
                std::mem::drop(img);

                let g = (0.299 * r as f32 + 0.587 * g as f32 + 0.114 * b as f32) as u8;
                let gc = Color32::from_rgba_unmultiplied(g, g, g, a);

                let mut img = arcci.write().unwrap();
                img[(x,y)] = gc;
            }
        }
    }
}

// From: https://docs.rs/egui/latest/egui/struct.ColorImage.html
fn load_image_from_path(path: &std::path::Path) -> Result<egui::ColorImage, image::ImageError> {
    let image = image::ImageReader::open(path)?.decode()?;
    let size = [image.width() as _, image.height() as _];
    let image_buffer = image.to_rgba8();
    let pixels = image_buffer.as_flat_samples();
    Ok(egui::ColorImage::from_rgba_unmultiplied(
        size,
        pixels.as_slice(),
    ))
}

#[allow(unused)]
fn load_image_from_memory(image_data: &[u8]) -> Result<egui::ColorImage, image::ImageError> {
    let image = image::load_from_memory(image_data)?;
    let size = [image.width() as _, image.height() as _];
    let image_buffer = image.to_rgba8();
    let pixels = image_buffer.as_flat_samples();
    Ok(egui::ColorImage::from_rgba_unmultiplied(
        size,
        pixels.as_slice(),
    ))
}
